"""

File: inspy_logger/__about__.py
Project: inSPy-Logger
Description: 

Created: 11/3/22 - 21:05:06

Since:
    v2.1.1

"""

__prog__ = 'inSPy-Logger'
__version__ = '2.1.1'
__authors__ = [
        ('Inspyre-Softworks', 'https://inspyre.tech'),
        ('Taylor-Jayde Blackstone', '<t.blackstone@inspyre.tech')
]
